/*
 *	vloopback.c
 *
 *	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2000
 *
 *	Published under the GNU Public License.
 *
 *	The Video Loopback Device is no longer systematically maintained.
 *	The project is a secondary project for the project "motion" found
 *	at http://motion.sourceforge.net/ and with the vloopback stored
 *	at http://motion.sourceforge.net/vloopback
 *
 *
 *	UPDATED:	Jeroen Vreeken.
 *			Added locks for smp machines. UNTESTED!
 *			Made the driver much more cpu friendly by using
 *			a wait queue.
 *			Went from vmalloc to rvmalloc (yes, I stole the code
 *			like everybody else) and implemented mmap.
 *			Implemented VIDIOCGUNIT and removed size/palette checks
 *			in VIDIOCSYNC.
 *			Cleaned up a lot of code.
 *			Changed locks to semaphores.
 *			Disabled changing size while somebody is using mmap
 *			Changed mapped check to open check, also don't allow
 *			a open for write while somebody is reading.
 *			Added /proc support
 *			Set dumped count to zero at open.
 *			Modified /proc layout (added vloopbacks entry)
 *
 * 05.10.00 (MTS)	Added Linux 2.2 support
 * 06.10.00 (J Vreeken)	Fixed 2.2 support to make things work under 2.4 again.
 * 17.10.00 (J Vreeken)	Added zero copy mode
 * 19.10.00 (J Vreeken) Added SIGIO on device close.
 * 24.10.00 (J Vreeken) Modified 2.2 stuff and removed spinlock.h
 *			released 0.81
 * 27.10.00 (J Vreeken) Implemented poll
 *			released 0.82
 * 17.01.01 (J Vreeken) support for xawtv
 * 			Implemented VIDIOCGFBUF
 *			Additional checks on framebuffer freeing.
 *			released 0.83
 * 31.01.01 (J Vreeken)	Removed need for 'struct ioctl', use _IOC_SIZE() and
 *			IOC_IN instead.
 *			Change the ioctlnr passing to 'unsigned long int'
 *			Instead of just one byte.
 *			THIS BREAKS COMPATIBILITY WITH PREVIOUS VERSIONS!!! 
 * 29.06.01 (J Vreeken)	Added dev_offset module option
 *			Made vloopback_template sane
 *			Added double buffering support
 *			Made vloopback less verbose
 * 20.11.01	(tibit)	Made dev_offset option sane
 *			"Fixed" zerocopy mode by defining the ioctl 
 *			VIDIOCSINVALID. An application which provides data
 *			has to issue it when it encounters an error in
 *			ioctl processing. See dummy.c for examples.
 * 26.11.03	(Kenneth Lavrsen)
 *			released 0.91
 *			0.91 is the combination of the 0.90-tibit by
 *			Tilmann Bitterberg and an update of the Makefile by
 *			Roberto Carvajal.
 * 13.04.04	(Kenneth Lavrsen) v. 0.92
 *			Applied patch for kernel using the new rmap-vm by
 *			Angel Carpintero
 */


#define VLVER "0.92"

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/wrapper.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/videodev.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/ioctl.h>

#ifndef HAVE_VMA
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,5,3) || defined(pte_offset_map)
#define HAVE_VMA 1
#else
#define HAVE_VMA 0
#endif
#endif

#if HAVE_VMA
#define pte_offset(pmd,adr) pte_offset_map(pmd,adr)     /* Emulation for a kernel using the new rmap-vm */
#endif

#define VIDIOCSINVALID	_IO('v',BASE_VIDIOCPRIVATE+1)

/************************************************************
    2.2 support
 ************************************************************/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
# include <linux/vmalloc.h>

/* define some helpers */
//typedef struct wait_queue* wait_queue_head_t;

# define virt_to_page(arg)         MAP_NR(arg)
# define page_address(arg)         arg
# define vmalloc_32                vmalloc

//# define init_waitqueue_head(arg)  init_waitqueue(arg)
//# define init_MUTEX(arg)           *(arg) = MUTEX
#endif
/************************************************************/

#define info(format, arg...) printk(KERN_INFO __FILE__ ": " format "\n" "", ## arg)

struct vloopback_device {
	struct video_device viddev;
	int pipenr;
	int in; /* bool */
};

struct vloopback_pipe {
	struct vloopback_device *vloopin;
 	struct vloopback_device *vloopout;
	char *buffer;
	unsigned long buflength;
	unsigned int width, height;
	unsigned int palette;
	unsigned long frameswrite;
	unsigned long framesread;
	unsigned long framesdumped;
	unsigned int wopen;
	unsigned int ropen;
	struct proc_dir_entry *proc_entry;
	struct semaphore lock;
	wait_queue_head_t wait;
	unsigned int frame;
	unsigned int pid;
	unsigned int zerocopy;
	unsigned long int ioctlnr;
	unsigned int invalid_ioctl; /* 0 .. none invalid; 1 .. invalid */
	unsigned int ioctllength;
	char *ioctldata;
	char *ioctlretdata;
};

#define MAX_PIPES 16

static struct vloopback_pipe *loops[MAX_PIPES];
static int nr_o_pipes=0;
static int pipes=-1;
static int spares=0;
static int pipesused=0;
static int dev_offset=-1;

/**********************************************************************
 *
 * Memory management
 *
 * This is a shameless copy from the USB-cpia driver (linux kernel
 * version 2.3.29 or so, I have no idea what this code actually does ;).
 * Actually it seems to be a copy of a shameless copy of the bttv-driver.
 * Or that is a copy of a shameless copy of ... (To the powers: is there
 * no generic kernel-function to do this sort of stuff?)
 *
 * Yes, it was a shameless copy from the bttv-driver. IIRC, Alan says
 * there will be one, but apparentely not yet -jerdfelt
 *
 * So I copied it again for the OV511 driver -claudio
 *
 * And it gets copied and copied and copied..... -Jeroen
 *
 **********************************************************************/

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
        unsigned long ret = 0UL;
        pmd_t *pmd;
        pte_t *ptep, pte;

        if (!pgd_none(*pgd)) {
                pmd = pmd_offset(pgd, adr);
                if (!pmd_none(*pmd)) {
                        ptep = pte_offset(pmd, adr);
                        pte = *ptep;
                        if (pte_present(pte)) {
                                ret = (unsigned long) page_address(pte_page(pte));
                                ret |= (adr & (PAGE_SIZE - 1));
                        }
                }
        }

        return ret;
}

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the
 * area and marking the pages as reserved.
 */
static inline unsigned long kvirt_to_pa(unsigned long adr)
{
        unsigned long va, kva, ret;

        va = VMALLOC_VMADDR(adr);
        kva = uvirt_to_kva(pgd_offset_k(va), va);
        ret = __pa(kva);
        return ret;
}

static void *rvmalloc(unsigned long size)
{
        void *mem;
        unsigned long adr, page;

        /* Round it off to PAGE_SIZE */
        size += (PAGE_SIZE - 1);
        size &= ~(PAGE_SIZE - 1);

        mem = vmalloc_32(size);
        if (!mem)
                return NULL;

        memset(mem, 0, size); /* Clear the ram out, no junk to the user */
        adr = (unsigned long) mem;
        while (size > 0) {
                page = kvirt_to_pa(adr);
                mem_map_reserve(virt_to_page(__va(page)));
                adr += PAGE_SIZE;
                if (size > PAGE_SIZE)
                        size -= PAGE_SIZE;
                else
                        size = 0;
        }

        return mem;
}

static void rvfree(void *mem, unsigned long size)
{
        unsigned long adr, page;

        if (!mem)
                return;

        size += (PAGE_SIZE - 1);
        size &= ~(PAGE_SIZE - 1);

        adr=(unsigned long) mem;
        while (size > 0) {
                page = kvirt_to_pa(adr);
                mem_map_unreserve(virt_to_page(__va(page)));
                adr += PAGE_SIZE;
                if (size > PAGE_SIZE)
                        size -= PAGE_SIZE;
                else
                        size = 0;
        }
        vfree(mem);
}

/**********************************************************************
 *
 * End of memory stuff, the rest is mine :)
 *
 **********************************************************************/
static int create_pipe(int nr);

static int fake_ioctl(int nr, unsigned long int cmd, void *arg)
{
	loops[nr]->ioctlnr=cmd;
	memcpy(loops[nr]->ioctldata, arg, _IOC_SIZE(cmd));
	loops[nr]->ioctllength=_IOC_SIZE(cmd);
	kill_proc(loops[nr]->pid, SIGIO, 1);
	interruptible_sleep_on(&loops[nr]->wait);
	if (cmd & IOC_IN) {
		if (memcmp (arg, loops[nr]->ioctlretdata, _IOC_SIZE(cmd)))
			return 1;
	} else {
		memcpy (arg, loops[nr]->ioctlretdata, _IOC_SIZE(cmd));
	}
	return 0;
}

static int vloopback_open(struct video_device *dev, int flags)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)dev;
	int nr=loopdev->pipenr;

	/* Only allow a output to be opened if there is someone feeding
	 * the pipe.
	 */
	if (!loopdev->in) {
		if (loops[nr]->buffer==NULL) {
			return -EINVAL;
		}
		loops[nr]->framesread=0;
		loops[nr]->ropen=1;
	} else {
		if (loops[nr]->ropen)
			return -EBUSY;
		loops[nr]->framesdumped=0;
		loops[nr]->frameswrite=0;
		loops[nr]->wopen=1;
		loops[nr]->zerocopy=0;
		loops[nr]->ioctlnr=-1;
		pipesused++;
		if (nr_o_pipes-pipesused<spares) {
			if (!create_pipe(nr_o_pipes)) {
				info("Creating extra spare pipe");
				info("Loopback %d registered, input: video%d, output: video%d",
				    nr_o_pipes,
				    loops[nr_o_pipes]->vloopin->viddev.minor,
				    loops[nr_o_pipes]->vloopout->viddev.minor
				);
				nr_o_pipes++;
			}
		}
		loops[nr]->pid=current->pid;
	}
	MOD_INC_USE_COUNT;
	return 0;
}

static void vloopback_close(struct video_device *dev)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)dev;
	int nr=loopdev->pipenr;
	
	if (loopdev->in) {
		down(&loops[nr]->lock);
		if (loops[nr]->buffer && !loops[nr]->ropen) {
			rvfree(loops[nr]->buffer,
			    loops[nr]->buflength*2);
			loops[nr]->buffer=NULL;
		}
		up(&loops[nr]->lock);
		if (waitqueue_active(&loops[nr]->wait))
			wake_up(&loops[nr]->wait);

		loops[nr]->width=0;
		loops[nr]->height=0;
		loops[nr]->palette=0;
		loops[nr]->wopen=0;
		pipesused--;
	} else {
		down(&loops[nr]->lock);
		if (loops[nr]->buffer && !loops[nr]->wopen) {
			rvfree(loops[nr]->buffer,
			    loops[nr]->buflength*3);
			loops[nr]->buffer=NULL;
		}
		up(&loops[nr]->lock);
		loops[nr]->ropen=0;
		if (loops[nr]->zerocopy && loops[nr]->buffer) {
			loops[nr]->ioctlnr=0;
			loops[nr]->ioctllength=0;
			kill_proc(loops[nr]->pid, SIGIO, 1);
		}
	}

	MOD_DEC_USE_COUNT;
}

static int vloopback_init_done(struct video_device *dev)
{
	return 0;
}

static long vloopback_write(struct video_device *v, const char *buf, unsigned long count, int noblock)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)v;
	int nr=loopdev->pipenr;
	unsigned long realcount=count;
	
	if (!loopdev->in)
		return -EINVAL;
	if (loops[nr]->zerocopy)
		return -EINVAL;
	
	if (loops[nr]->buffer==NULL) {
		return -EINVAL;
	}

	/* Anybody want some pictures??? */
	if (!waitqueue_active(&loops[nr]->wait)) {
		loops[nr]->framesdumped++;
		return realcount;
	}
	
	down(&loops[nr]->lock);
	if (!loops[nr]->buffer) {
		up(&loops[nr]->lock);
		return -EINVAL;
	}
	if (realcount > loops[nr]->buflength) {
		realcount = loops[nr]->buflength;
		info("To much data! Only %ld bytes used.", realcount);
	}
	
	copy_from_user(
	    loops[nr]->buffer+loops[nr]->frame*loops[nr]->buflength,
	    buf, realcount
	);
	loops[nr]->frame=0;
	up(&loops[nr]->lock);

	wake_up(&loops[nr]->wait);
	loops[nr]->frameswrite++;

	return realcount;
}

static long vloopback_read(struct video_device *v, char *buf, unsigned long count, int noblock)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)v;
	int nr=loopdev->pipenr;
	unsigned long realcount=count;

	if (loops[nr]->zerocopy) {
		if (loopdev->in) {
			//info("DEBUG: vl_read: ->in: pipenr: %d", nr);
			if (realcount > loops[nr]->ioctllength+sizeof(unsigned long int))
				realcount=loops[nr]->ioctllength+sizeof(unsigned long int);
			//info("DEBUG: vl_read: ->in: realcount: %lu", realcount);
			copy_to_user(buf , &loops[nr]->ioctlnr, sizeof(unsigned long int));
			//info("DEBUG: vl_read: ->in: ioctlnr: %lu", loops[nr]->ioctlnr);
			copy_to_user(buf+sizeof(unsigned long int) , loops[nr]->ioctldata, 
					realcount-sizeof(unsigned long int));
			if (loops[nr]->ioctlnr==0)
				loops[nr]->ioctlnr=-1;
			return realcount;
		} else {
			struct video_window vidwin;
			struct video_mmap vidmmap;
			struct video_picture vidpic;
			//info("DEBUG: vl_read: ->out: pipenr: %d", nr);
			
			fake_ioctl(nr, VIDIOCGWIN, &vidwin);
			fake_ioctl(nr, VIDIOCGPICT, &vidpic);

			vidmmap.height=vidwin.height;
			vidmmap.width=vidwin.width;
			vidmmap.format=vidpic.palette;
			vidmmap.frame=0;
			if (fake_ioctl(nr, VIDIOCMCAPTURE, &vidmmap))
				return 0;
			if (fake_ioctl(nr, VIDIOCSYNC, &vidmmap))
				return 0;
			realcount=vidwin.height*vidwin.width*vidpic.depth;
		}
	}
	if (loopdev->in)
		return -EINVAL;

	if (realcount > loops[nr]->buflength) {
		realcount = loops[nr]->buflength;
		info("Not so much data in buffer!");
	}

	if (!loops[nr]->zerocopy)
		interruptible_sleep_on(&loops[nr]->wait);

	down(&loops[nr]->lock);
	if (!loops[nr]->buffer) {
		up(&loops[nr]->lock);
		return 0;
	}
	copy_to_user(buf, loops[nr]->buffer, realcount);	
	up(&loops[nr]->lock);

	loops[nr]->framesread++;
	return realcount;
}

#if HAVE_VMA
static int vloopback_mmap(struct vm_area_struct *vma,struct video_device *dev, const char *adr, unsigned long size)
#else
static int vloopback_mmap(struct video_device *dev, const char *adr, unsigned long size)
#endif
{
	struct vloopback_device *loopdev=(struct vloopback_device *)dev;
	int nr=loopdev->pipenr;
        unsigned long start = (unsigned long)adr;
        unsigned long page, pos;

	down(&loops[nr]->lock);
	//info("DEBUG: vloopback_mmap: Enter");
	if (loopdev->in) {
		/* Zero copy mode */
		//info("DEBUG: vloopback_mmap: Zero copy mode");
		loops[nr]->zerocopy=1;
		if (loops[nr]->ropen) {
			info("Can't change size while opened for read");
			up(&loops[nr]->lock);
			return -EINVAL;
		}
		if (!size) {
			up(&loops[nr]->lock);
			return -EINVAL;
		}
		if (loops[nr]->buffer)
			rvfree(loops[nr]->buffer, loops[nr]->buflength*2);
		loops[nr]->buflength=size;
		loops[nr]->buffer=rvmalloc(loops[nr]->buflength*2);
	}
        if (loops[nr]->buffer == NULL) {
		up(&loops[nr]->lock);
                return -EINVAL;
	}

        if (size > (((2 * loops[nr]->buflength) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))) {
		up(&loops[nr]->lock);
                return -EINVAL;
	}

        pos = (unsigned long)loops[nr]->buffer;
        while (size > 0) {
                page = kvirt_to_pa(pos);
#if HAVE_VMA
                if (remap_page_range(vma,start, page, PAGE_SIZE, PAGE_SHARED)) {
#else
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED)) {
#endif
			up(&loops[nr]->lock);
                        return -EAGAIN;
		}
                start += PAGE_SIZE;
                pos += PAGE_SIZE;
                if (size > PAGE_SIZE)
                        size -= PAGE_SIZE;
                else
                        size = 0;
        }
	up(&loops[nr]->lock);

	return 0;
}

static int vloopback_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)dev;
	int nr=loopdev->pipenr;
	int i;

	if (loops[nr]->zerocopy) {
		if (!loopdev->in) {
			loops[nr]->ioctlnr=cmd;
			loops[nr]->ioctllength=_IOC_SIZE(cmd);
			/* info("DEBUG: vl_ioctl: !loop->in"); */
			/* info("DEBUG: vl_ioctl: cmd %lu", cmd); */
			/* info("DEBUG: vl_ioctl: len %lu", loops[nr]->ioctllength); */
			if(copy_from_user(loops[nr]->ioctldata, arg, _IOC_SIZE(cmd)))
				return -EFAULT;
			kill_proc(loops[nr]->pid, SIGIO, 1);
			interruptible_sleep_on(&loops[nr]->wait);
			
			if (loops[nr]->invalid_ioctl) {
				//info ("DEBUG: There was an invalid ioctl");
				loops[nr]->invalid_ioctl = 0;
				return -ENOTTY;
			}
			if (cmd & IOC_IN && !(cmd & IOC_OUT)) {
				//info("DEBUG: vl_ioctl: cmd & IOC_IN 1"); 
				if (memcmp(loops[nr]->ioctlretdata, loops[nr]->ioctldata, _IOC_SIZE(cmd))) {
					return -EINVAL;
				}
			 	//info("DEBUG: vl_ioctl: cmd & IOC_IN 2"); 
				return 0;
			} else {
				copy_to_user(arg, loops[nr]->ioctlretdata, _IOC_SIZE(cmd));
				//info("DEBUG: vl_ioctl: !(cmd & IOC_IN) 1");
				return 0;
			}
		} else {
			if ( (loops[nr]->ioctlnr!=cmd) && (cmd != (VIDIOCSINVALID))) {
				/* wrong ioctl */
				info("DEBUG: vo_ioctl: Wrong IOCTL");
				return 0;
			}
			if (cmd == VIDIOCSINVALID) {
				loops[nr]->invalid_ioctl = 1;
			} else {
				if (copy_from_user(loops[nr]->ioctlretdata, arg, loops[nr]->ioctllength))
					return -EFAULT;
				//info("DEBUG: vl_ioctl: loopdev->in: 1"); 
			}
			if (waitqueue_active(&loops[nr]->wait))
				wake_up(&loops[nr]->wait);
			//info("DEBUG: vl_ioctl: loopdev->in: 2");
			loops[nr]->ioctlnr=-1;
			return 0;
		}
	}
	switch(cmd)
	{
		/* Get capabilities */
		case VIDIOCGCAP:
		{
			struct video_capability b;
			if (loopdev->in) {
				sprintf(b.name, "Video loopback %d input",
				    loopdev->pipenr);
				b.type = 0;
			} else {
				sprintf(b.name, "Video loopback %d output",
				    loopdev->pipenr);
				b.type = VID_TYPE_CAPTURE;
			}
			b.channels=1;
			b.audios=0;
			b.maxwidth=loops[nr]->width;
			b.maxheight=loops[nr]->height;
			b.minwidth=20;
			b.minheight=20;
			if(copy_to_user(arg, &b, sizeof(b)))
				return -EFAULT;
			return 0;
		}
		/* Get channel info (sources) */
		case VIDIOCGCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v.channel!=0) {
				info("VIDIOCGCHAN: Invalid Channel, was %d", v.channel);
				v.channel=0;
				//return -EINVAL;
			}
			v.flags=0;
			v.tuners=0;
			v.norm=0;
			v.type = VIDEO_TYPE_CAMERA;
			/*strcpy(v.name, "Loopback"); -- tibit */
			strcpy(v.name, "Composite1");
			if(copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}
		/* Set channel 	*/
		case VIDIOCSCHAN:
		{
			int v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			if(v!=0) {
				info("VIDIOCSCHAN: Invalid Channel, was %d", v);
				return -EINVAL;
			}
			return 0;
		}
		/* Get tuner abilities */
		case VIDIOCGTUNER:
		{
			struct video_tuner v;
			if(copy_from_user(&v, arg, sizeof(v))!=0)
				return -EFAULT;
			if(v.tuner) {
				info("VIDIOCGTUNER: Invalid Tuner, was %d", v.tuner);
				return -EINVAL;
			}
			strcpy(v.name, "Format");
			v.rangelow=0;
			v.rangehigh=0;
			v.flags=0;
			v.mode=VIDEO_MODE_AUTO;
			if(copy_to_user(arg,&v, sizeof(v))!=0)
				return -EFAULT;
			return 0;
		}
		/* Get picture properties */
		case VIDIOCGPICT:
		{
			struct video_picture p;
			p.colour=0x8000;
			p.hue=0x8000;
			p.brightness=0x8000;
			p.contrast=0x8000;
			p.whiteness=0x8000;
			p.depth=0x8000;
			p.palette=loops[nr]->palette;
			if(copy_to_user(arg, &p, sizeof(p)))
				return -EFAULT;
			return 0;

		}
		/* Set picture properties */
		case VIDIOCSPICT:
		{
			struct video_picture p;
			if(copy_from_user(&p, arg, sizeof(p)))
				return -EFAULT;
			if (!loopdev->in) {
				if (p.palette!=loops[nr]->palette)
					return -EINVAL;
			} else
				loops[nr]->palette=p.palette;
			return 0;
		}
		/* Get the video overlay window */
		case VIDIOCGWIN:
		{
			struct video_window vw;
			vw.x=0;
			vw.y=0;
			vw.width=loops[nr]->width;
			vw.height=loops[nr]->height;
			vw.chromakey=0;
			vw.flags=0;
			vw.clipcount=0;
			if(copy_to_user(arg, &vw, sizeof(vw)))
				return -EFAULT;
			return 0;
		}
		/* Set the video overlay window - passes clip list for hardware smarts , chromakey etc */
		case VIDIOCSWIN:
		{
			struct video_window vw;
			
			if(copy_from_user(&vw, arg, sizeof(vw)))
				return -EFAULT;
			if(vw.flags)
				return -EINVAL;
			if(vw.clipcount)
				return -EINVAL;
			if (loops[nr]->height==vw.height &&
			    loops[nr]->width==vw.width)
				return 0;
			if(!loopdev->in) {
				return -EINVAL;
			} else {
				loops[nr]->height=vw.height;
				loops[nr]->width=vw.width;
				/* Make sure nobody is using the buffer while we
				   fool around with it.
				   We are also not allowing changes while
				   somebody using mmap has the output open.
				 */
				down(&loops[nr]->lock);
				if (loops[nr]->ropen) {
					info("Can't change size while opened for read");
					up(&loops[nr]->lock);
					return -EINVAL;
				}
				if (loops[nr]->buffer)
					rvfree(loops[nr]->buffer, loops[nr]->buflength*2);
				loops[nr]->buflength=vw.width*vw.height*4;
				loops[nr]->buffer=rvmalloc(loops[nr]->buflength*2);
				up(&loops[nr]->lock);
			}
			return 0;
		}
		/* Memory map buffer info */
		case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			
			vm.size=loops[nr]->buflength*2;
			vm.frames=2;
			for (i=0; i<vm.frames; i++)
				vm.offsets[i]=i*loops[nr]->buflength;
			if(copy_to_user(arg, &vm, sizeof(vm)))
				return -EFAULT;
			return 0;
		}
		/* Grab frames */
		case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;

			if (loopdev->in)
				return -EINVAL;
			if (!loops[nr]->buffer)
				return -EINVAL;
			if (copy_from_user(&vm, arg, sizeof(vm)))
				return -EFAULT;
			if (vm.format!=loops[nr]->palette)
				return -EINVAL;
			if (vm.frame > 2)
				return -EINVAL;
			/* Dirty xawtv hack :( */
			/* is this needed anymore? 
			if (vm.width > 64 || vm.height > 64)
				if (vm.height!=loops[nr]->height ||
				    vm.width!=loops[nr]->width)
					return -EINVAL;
			 	*/
			return 0;
		}
		/* Sync with mmap grabbing */
		case VIDIOCSYNC:
		{
			int frame;

			if (copy_from_user((void *)&frame, arg, sizeof(int)))
				return -EFAULT;
			if (loopdev->in)
				return -EINVAL;
			if (!loops[nr]->buffer)
				return -EINVAL;
			/* Ok, everything should be alright since the program
			   should have called VIDIOMCAPTURE and we are ready to
			   do the 'capturing' */
			if (frame > 1)
				return -EINVAL;
			loops[nr]->frame=frame;
			interruptible_sleep_on(&loops[nr]->wait);
			loops[nr]->framesread++;
			return 0;
		}
		/* Get attached units */
		case VIDIOCGUNIT:
		{
			struct video_unit vu;
			
			if (loopdev->in)
				vu.video=loops[nr]->vloopout->viddev.minor;
			else
				vu.video=loops[nr]->vloopin->viddev.minor;
			vu.vbi=VIDEO_NO_UNIT;
			vu.radio=VIDEO_NO_UNIT;
			vu.audio=VIDEO_NO_UNIT;
			vu.teletext=VIDEO_NO_UNIT;
			if (copy_to_user(arg, &vu, sizeof(vu)))
				return -EFAULT;
			return 0;
		}
		/* Get frame buffer */
		case VIDIOCGFBUF:
		{
			struct video_buffer vb;

			memset(&vb, 0, sizeof(vb));
			vb.base=NULL;

			if(copy_to_user((void *)arg, (void *)&vb, sizeof(vb)))
				return -EFAULT;

			return 0;
		}
		/* Start, end capture */
		case VIDIOCCAPTURE:
		{
			int start;
			if (copy_from_user(&start, arg, sizeof(int)))
				return -EFAULT;
			if (start) info ("Capture started");
			else info ("Capture stopped");

			return 0;
		}

		case VIDIOCGFREQ:
		case VIDIOCSFREQ:
		case VIDIOCGAUDIO:
		case VIDIOCSAUDIO:
			return -EINVAL;
		case VIDIOCKEY:
			return 0;
		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}

static unsigned int vloopback_poll(struct video_device *dev, struct file *file, poll_table *wait)
{
	struct vloopback_device *loopdev=(struct vloopback_device *)dev;
	int nr=loopdev->pipenr;

	if (dev==NULL)
		return -EFAULT;
	if (!loopdev->in)
		return 0;

	if (loops[nr]->ioctlnr!=-1) {
		if (loops[nr]->zerocopy) {
			return (POLLIN | POLLPRI | POLLOUT | POLLRDNORM);
		} else {
			return (POLLOUT);
		}
	}
	return 0;
}


static struct video_device vloopback_template=
{
	owner:		THIS_MODULE,
	name:		"Video Loopback",
	type:		VID_TYPE_CAPTURE,
	open:		vloopback_open,
	close:		vloopback_close,
	read:		vloopback_read,
	write:		vloopback_write,
	poll:		vloopback_poll,
	ioctl:		vloopback_ioctl,
	mmap:		vloopback_mmap,
	initialize:	vloopback_init_done,
};

/****************************************************************************
 *	/proc interface
 *	Based on the ov511 driver
 ****************************************************************************/

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)

static struct proc_dir_entry *vloopback_proc_entry = NULL;
static struct proc_dir_entry *vloopbacks_proc_entry = NULL;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
extern struct proc_dir_entry *video_proc_entry;
#else
static struct proc_dir_entry *video_proc_entry = NULL;
/* /proc/video is created here */
#endif

#define YES_NO(x) ((x) ? "yes" : "no")
#define STAT_W(x) ((x) ? "W" : "-")
#define STAT_R(x) ((x) ? "R" : "-")

static int vloopbacks_read_proc(char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	char *out=page;
	int len, i;
	
	out += sprintf(out, "vloopback version "VLVER"\n");
	out += sprintf(out, "Loop\tInput\tStatus\tOutput\tStatus\n");
	for (i=0; i<nr_o_pipes; i++) {
		out += sprintf(out, "%d\tvideo%d\t%s\tvideo%d\t%s\n", i,
			       loops[i]->vloopin->viddev.minor,
			       STAT_W(loops[i]->wopen),
			       loops[i]->vloopout->viddev.minor,
			       STAT_R(loops[i]->ropen)
			       );
	}

	len = out - page;
	len -= off;
	if (len < count) {
		*eof=1;
		if (len <= 0) return 0;
	} else
		len = count;
	
	*start = page + off;
	
	return len; 
}

static int vloopback_read_proc(char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	char *out=page;
	int len;
	struct vloopback_pipe *pipe=data;

	out += sprintf (out, "video%d\t%s\n", pipe->vloopin->viddev.minor, STAT_W(pipe->wopen));
	out += sprintf (out, "video%d\t%s\n", pipe->vloopout->viddev.minor, STAT_R(pipe->ropen));
	out += sprintf (out, "written\t%ld\n", pipe->frameswrite);
	out += sprintf (out, "dumped\t%ld\n", pipe->framesdumped);
	out += sprintf (out, "read\t%ld\n", pipe->framesread);

	len = out - page;
	len -= off;
	if (len < count) {
		*eof=1;
		if (len <= 0) return 0;
	} else
		len = count;
	
	*start = page + off;
	
	return len;
}

static int vloopback_write_proc(struct file *file, const char *buffer,
			      unsigned long count, void *data)
{
	return -EINVAL;
}

static void create_proc_vloopback (int nr)
{
	char name[16];
	struct proc_dir_entry *ent;
	
	if (!vloopback_proc_entry)
		return;

	sprintf(name, "vloopback%d", nr);
	
	ent=create_proc_entry(name, S_IFREG | S_IRUGO | S_IWUSR, vloopback_proc_entry);
	
	if (!ent) {
		loops[nr]->proc_entry=NULL;
		return;
	}

	ent->data = loops[nr];
	ent->read_proc = vloopback_read_proc;
	ent->write_proc = vloopback_write_proc;
	loops[nr]->proc_entry = ent;
}

static void destroy_proc_vloopback (int nr)
{
	char name[16];
	
	if (!loops[nr]->proc_entry)
		return;

	sprintf(name, "vloopback%d", nr);
	remove_proc_entry(name, vloopback_proc_entry);
	loops[nr]->proc_entry=NULL;
}

static void proc_vloopback_create(void)
{
	char name[16];
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
	video_proc_entry = create_proc_entry("video", S_IFDIR, 0);
#endif

	if (video_proc_entry == NULL) {
		info("Unable to initialise /proc/video/vloopback");
		return;
	}
	
	vloopback_proc_entry=create_proc_entry("vloopback", S_IFDIR, video_proc_entry);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
	if (vloopback_proc_entry)
		vloopback_proc_entry->owner = THIS_MODULE;
	else {
		info("Unable to initialise /proc/video/vloopback");
		return;
	}
#endif

	sprintf(name, "vloopbacks");
	
	vloopbacks_proc_entry=create_proc_entry(name, S_IFREG | S_IRUGO | S_IWUSR,
						vloopback_proc_entry);
	
	if (!vloopbacks_proc_entry) {
		return;
	}

	vloopbacks_proc_entry->data = NULL;
	vloopbacks_proc_entry->read_proc = vloopbacks_read_proc;
	vloopbacks_proc_entry->write_proc = vloopback_write_proc;
}

static void proc_vloopback_destroy(void)
{
	if (vloopback_proc_entry == NULL)
		return;
		
	remove_proc_entry("vloopback", video_proc_entry);
}
#endif /* CONFIG_PROC_FS && CONFIG_VIDEO_PROC_FS */


static int create_pipe(int nr)
{
	int minor_in, minor_out;
	if (dev_offset == -1)
		minor_in  = minor_out = -1; /* autoassign */
	else {
		minor_in  = 2*nr   + dev_offset;
		minor_out = 2*nr+1 + dev_offset;
	}

	loops[nr]= kmalloc(sizeof(struct vloopback_pipe), GFP_KERNEL);
	if (!loops[nr])
		return -ENOMEM;
	loops[nr]->vloopin= kmalloc(sizeof(struct vloopback_device), GFP_KERNEL);
	if (!loops[nr]->vloopin)
		return -ENOMEM;
	loops[nr]->vloopout= kmalloc(sizeof(struct vloopback_device), GFP_KERNEL);
	if (!loops[nr]->vloopout)
		return -ENOMEM;
	loops[nr]->vloopin->pipenr=nr;
	loops[nr]->vloopout->pipenr=nr;
	loops[nr]->invalid_ioctl = 0; /* tibit */
	loops[nr]->buffer=NULL;
	loops[nr]->width=0;
	loops[nr]->height=0;
	loops[nr]->palette=0;
	loops[nr]->frameswrite=0;
	loops[nr]->framesread=0;
	loops[nr]->framesdumped=0;
	loops[nr]->wopen=0;
	loops[nr]->ropen=0;
	loops[nr]->frame=0;
	memcpy(loops[nr]->vloopin, &vloopback_template, sizeof(vloopback_template));
	memcpy(loops[nr]->vloopout, &vloopback_template, sizeof(vloopback_template));
	loops[nr]->vloopin->in=1;
	loops[nr]->vloopout->in=0;
	((struct video_device *)loops[nr]->vloopin)->type=0;
	sprintf(((struct video_device *)loops[nr]->vloopin)->name, "Video loopback %d input", nr);
	((struct video_device *)loops[nr]->vloopout)->type=VID_TYPE_CAPTURE;
	sprintf(((struct video_device *)loops[nr]->vloopout)->name, "Video loopback %d output", nr);
	init_waitqueue_head(&loops[nr]->wait);
	init_MUTEX(&loops[nr]->lock);
	if (video_register_device(&loops[nr]->vloopin->viddev, VFL_TYPE_GRABBER, minor_in)==-1) {
	kfree(loops[nr]->vloopin);
		kfree(loops[nr]->vloopout);
		kfree(loops[nr]);
		loops[nr]=NULL;
		info("error registering device");
		return -ENODEV;
	}
	if (video_register_device(&loops[nr]->vloopout->viddev, VFL_TYPE_GRABBER, minor_out)==-1) {
		video_unregister_device(&loops[nr]->vloopin->viddev);
		kfree(loops[nr]->vloopin);
		kfree(loops[nr]->vloopout);
		kfree(loops[nr]);
		loops[nr]=NULL;
		info("error registering device");
		return -ENODEV;
	}
	loops[nr]->ioctldata=kmalloc(1024, GFP_KERNEL);
	loops[nr]->ioctlretdata=kmalloc(1024, GFP_KERNEL);
#if defined(CONFIG_PROC_FS) && defined (CONFIG_VIDEO_PROC_FS)
	create_proc_vloopback(nr);
#endif
	return 0;
}


/****************************************************************************
 *	init stuff
 ****************************************************************************/


MODULE_AUTHOR("J.B. Vreeken (pe1rxq@amsat.org)");
MODULE_DESCRIPTION("Video4linux loopback device.");
MODULE_PARM(pipes, "i");
MODULE_PARM_DESC(pipes, "Nr of pipes to create (each pipe uses two video devices)");
MODULE_PARM(spares, "i");
MODULE_PARM_DESC(spares, "Nr of spare pipes that should be created");
MODULE_PARM(dev_offset, "i");
MODULE_PARM_DESC(dev_offset, "Prefered offset for video device numbers");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 10)
MODULE_LICENSE("GPL");
#endif

EXPORT_NO_SYMBOLS;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
static int __init vloopback_init(void)
#else
int init_module(void)
#endif
{
	int i;

	info("Video4linux loopback driver v"VLVER);

	if (pipes==-1) pipes=1;
	if (pipes > MAX_PIPES) {
		pipes=MAX_PIPES;
		info("Nr of pipes is limited to: %d", MAX_PIPES);
	}
	
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	proc_vloopback_create();
#endif

	for (i=0; i<pipes; i++) {
		if (!create_pipe(i)) {
			info("Loopback %d registered, input: video%d, output: video%d",
			    i, loops[i]->vloopin->viddev.minor, loops[i]->vloopout->viddev.minor);
			nr_o_pipes=i+1;
		}
	}
	return 0;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
static void __exit cleanup_vloopback_module(void)
#else
void cleanup_module(void)
#endif
{
	int i;

	info("Unregistering video4linux loopback devices");
	for (i=0; i<nr_o_pipes; i++) if (loops[i]) {
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
		destroy_proc_vloopback(i);
#endif
		video_unregister_device(&loops[i]->vloopin->viddev);
		kfree(loops[i]->vloopin);
		video_unregister_device(&loops[i]->vloopout->viddev);
		kfree(loops[i]->vloopout);
		if (loops[i]->buffer) rvfree(loops[i]->buffer, loops[i]->buflength*2);
		kfree(loops[i]->ioctldata);
		kfree(loops[i]->ioctlretdata);
		kfree(loops[i]);
	}
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	proc_vloopback_destroy();
#endif
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 3, 0)
module_init(vloopback_init);
module_exit(cleanup_vloopback_module);
#endif
